# -*- coding: utf-8 -*-
"""
Modified by Group 04
"""

import random
import time
from halma_game.halma_player import HalmaPlayer
from halma_game.halma_model import HalmaModel


class HalmaPlayer01B(HalmaPlayer):
    def __init(self, nama):
        super().__init__(nama)